# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox module class"""
from dataclasses import dataclass
from typing import Sequence, cast

from qpysequence import Sequence as QpySequence

from qililab.config import logger
from qililab.instruments.awg import AWG
from qililab.instruments.awg_settings import AWGQbloxSequencer
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.pulse.pulse_bus_schedule import PulseBusSchedule
from qililab.typings.enums import Parameter
from qililab.typings.instruments import Pulsar, QcmQrm


class QbloxModule(AWG):
    """Qblox Module class.

    Args:
        device (Pulsar): Instance of the Qblox Pulsar class used to connect to the instrument.
        settings (QbloxPulsarSettings): Settings of the instrument.
    """

    _MAX_BINS: int = 131072
    _NUM_MAX_SEQUENCERS: int = 6
    _NUM_MAX_AWG_OUT_CHANNELS: int = 4
    _MIN_WAIT_TIME: int = 4  # in ns

    @dataclass
    class QbloxModuleSettings(AWG.AWGSettings):
        """Contains the settings of a specific pulsar.

        Args:
            awg_sequencers (Sequence[AWGQbloxSequencer]): list of settings for each sequencer
            out_offsets (list[float]): list of offsets for each output of the qblox module
        """

        awg_sequencers: Sequence[AWGQbloxSequencer]
        out_offsets: list[float]

        def __post_init__(self):
            """build AWGQbloxSequencer"""
            if (
                self.num_sequencers <= 0
                or self.num_sequencers > QbloxModule._NUM_MAX_SEQUENCERS  # pylint: disable=protected-access
            ):
                raise ValueError(
                    "The number of sequencers must be greater than 0 and less or equal than "
                    + f"{QbloxModule._NUM_MAX_SEQUENCERS}. Received: {self.num_sequencers}"  # pylint: disable=protected-access
                )
            if len(self.awg_sequencers) != self.num_sequencers:
                raise ValueError(
                    f"The number of sequencers: {self.num_sequencers} does not match"
                    + f" the number of AWG Sequencers settings specified: {len(self.awg_sequencers)}"
                )

            self.awg_sequencers = [
                AWGQbloxSequencer(**sequencer)
                if isinstance(sequencer, dict)
                else sequencer  # pylint: disable=not-a-mapping
                for sequencer in self.awg_sequencers
            ]
            super().__post_init__()

    settings: QbloxModuleSettings
    device: Pulsar | QcmQrm
    # Cache containing the last compiled pulse schedule for each sequencer
    cache: dict[int, PulseBusSchedule] = {}

    def __init__(self, settings: dict):
        # The sequences dictionary contains all the compiled sequences for each sequencer. Sequences are saved and handled at the compiler
        self.sequences: dict[int, QpySequence] = {}  # {sequencer_idx: (program), ...}
        self.num_bins: int = 1
        super().__init__(settings=settings)

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """Initial setup"""
        self._map_connections()
        self.clear_cache()
        for sequencer in self.awg_sequencers:
            sequencer_id = sequencer.identifier
            # Set `sync_en` flag to False (this value will be set to True if the sequencer is used in the execution)
            self.device.sequencers[sequencer_id].sync_en(False)
            self._set_nco(sequencer_id=sequencer_id)
            self._set_gain_i(value=sequencer.gain_i, sequencer_id=sequencer_id)
            self._set_gain_q(value=sequencer.gain_q, sequencer_id=sequencer_id)
            self._set_offset_i(value=sequencer.offset_i, sequencer_id=sequencer_id)
            self._set_offset_q(value=sequencer.offset_q, sequencer_id=sequencer_id)
            self._set_hardware_modulation(value=sequencer.hardware_modulation, sequencer_id=sequencer_id)
            self._set_gain_imbalance(value=sequencer.gain_imbalance, sequencer_id=sequencer_id)
            self._set_phase_imbalance(value=sequencer.phase_imbalance, sequencer_id=sequencer_id)
            ALL_ON = 15  # 1111 in binary
            self._set_markers(value=ALL_ON, sequencer_id=sequencer_id)

        for idx, offset in enumerate(self.out_offsets):
            self._set_out_offset(output=idx, value=offset)

    def sync_by_port(self, port: str) -> None:
        """Syncs all sequencers."""
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=port)
        for sequencer in sequencers:
            self.device.sequencers[sequencer.identifier].sync_en(True)

    def desync_by_port(self, port: str) -> None:
        """Syncs all sequencers."""
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=port)
        for sequencer in sequencers:
            self.device.sequencers[sequencer.identifier].sync_en(False)

    def desync_sequencers(self) -> None:
        """Desyncs all sequencers."""
        for sequencer in self.awg_sequencers:
            self.device.sequencers[sequencer.identifier].sync_en(False)

    @property
    def module_type(self):
        """returns the qblox module type. Options: QCM or QRM"""
        return self.device.module_type()

    def run(self, port: str):
        """Run the uploaded program"""
        self.start_sequencer(port=port)

    def start_sequencer(self, port: str):
        """Start sequencer and execute the uploaded instructions."""
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=port)
        for sequencer in sequencers:
            if sequencer.identifier in self.sequences:
                self.device.arm_sequencer(sequencer=sequencer.identifier)
                self.device.start_sequencer(sequencer=sequencer.identifier)

    def setup(  # pylint: disable=too-many-branches, too-many-return-statements
        self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None
    ):
        """Set Qblox instrument calibration settings."""
        if parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            self._set_out_offset(output=output, value=value)
            return

        if channel_id is None:
            if self.num_sequencers == 1:
                channel_id = 0
            else:
                raise ParameterNotFound(f"Cannot update parameter {parameter.value} without specifying a channel_id.")

        if channel_id > self.num_sequencers - 1:
            raise ParameterNotFound(
                f"the specified channel id:{channel_id} is out of range. Number of sequencers is {self.num_sequencers}"
            )
        if parameter == Parameter.GAIN:
            self._set_gain(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_I:
            self._set_gain_i(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_Q:
            self._set_gain_q(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.OFFSET_I:
            self._set_offset_i(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.OFFSET_Q:
            self._set_offset_q(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.IF:
            self._set_frequency(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.HARDWARE_MODULATION:
            self._set_hardware_modulation(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.NUM_BINS:
            self._set_num_bins(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_IMBALANCE:
            self._set_gain_imbalance(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.PHASE_IMBALANCE:
            self._set_phase_imbalance(value=value, sequencer_id=channel_id)
            return
        raise ParameterNotFound(f"Invalid Parameter: {parameter.value}")

    def get(self, parameter: Parameter, channel_id: int | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            return self.out_offsets[output]

        if channel_id is None:
            if self.num_sequencers == 1:
                channel_id = 0
            else:
                raise ParameterNotFound(f"Cannot update parameter {parameter.value} without specifying a channel_id.")

        sequencer = self._get_sequencer_by_id(id=channel_id)

        if parameter == Parameter.GAIN:
            return sequencer.gain_i, sequencer.gain_q

        if hasattr(sequencer, parameter.value):
            return getattr(sequencer, parameter.value)

        raise ParameterNotFound(f"Cannot find parameter {parameter.value} in instrument {self.alias}")

    @Instrument.CheckParameterValueFloatOrInt
    def _set_num_bins(self, value: float | str | bool, sequencer_id: int):
        """set num_bins for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        if int(value) > self._MAX_BINS:
            raise ValueError(f"Value {value} greater than maximum bins: {self._MAX_BINS}")
        cast(AWGQbloxSequencer, self._get_sequencer_by_id(id=sequencer_id)).num_bins = int(value)

    @Instrument.CheckParameterValueBool
    def _set_hardware_modulation(self, value: float | str | bool, sequencer_id: int):
        """set hardware modulation

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        self._get_sequencer_by_id(id=sequencer_id).hardware_modulation = bool(value)

        if self.is_device_active():
            self.device.sequencers[sequencer_id].mod_en_awg(bool(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_frequency(self, value: float | str | bool, sequencer_id: int):
        """set frequency

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self._get_sequencer_by_id(id=sequencer_id).intermediate_frequency = float(value)

        if self.is_device_active():
            self.device.sequencers[sequencer_id].nco_freq(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_offset_i(self, value: float | str | bool, sequencer_id: int):
        """Set the offset of the I channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).offset_i = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "offset_awg_path0")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_offset_q(self, value: float | str | bool, sequencer_id: int):
        """Set the offset of the Q channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).offset_q = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "offset_awg_path1")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_out_offset(self, output: int, value: float | str | bool):
        """Set output offsets of the Qblox device.

        Args:
            output (int): output to update
            value (float | str | bool): value to update

        Raises:
            ValueError: when value type is not float or int
        """
        if output > len(self.out_offsets):
            raise IndexError(
                f"Output {output} is out of range. The runcard has only {len(self.out_offsets)} output offsets defined."
                " Please update the list of output offsets of the runcard such that it contains a value for each "
                "output of the device."
            )
        self.out_offsets[output] = value

        if self.is_device_active():
            getattr(self.device, f"out{output}_offset")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain_i(self, value: float | str | bool, sequencer_id: int):
        """Set the gain of the I channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).gain_i = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "gain_awg_path0")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain_q(self, value: float | str | bool, sequencer_id: int):
        """Set the gain of the Q channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).gain_q = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "gain_awg_path1")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain(self, value: float | str | bool, sequencer_id: int):
        """set gain

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self._set_gain_i(value=value, sequencer_id=sequencer_id)
        self._set_gain_q(value=value, sequencer_id=sequencer_id)

    @Instrument.CheckDeviceInitialized
    def turn_off(self):
        """Stop the QBlox sequencer from sending pulses."""
        for seq_idx in range(self.num_sequencers):
            self.device.stop_sequencer(sequencer=seq_idx)

    @Instrument.CheckDeviceInitialized
    def turn_on(self):
        """Turn on an instrument."""

    def clear_cache(self):
        """Empty cache."""
        self.cache = {}
        self.sequences = {}

    @Instrument.CheckDeviceInitialized
    def reset(self):
        """Reset instrument."""
        self.clear_cache()
        self.device.reset()

    def upload_qpysequence(self, qpysequence: QpySequence, port: str):
        """Upload the qpysequence to its corresponding sequencer.

        Args:
            qpysequence (QpySequence): The qpysequence to upload.
            port (str): The port of the sequencer to upload to.
        """
        # FIXME: does not support readout on multiple qubits on the same bus
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=port)
        for sequencer in sequencers:
            logger.info("Sequence program: \n %s", repr(qpysequence._program))  # pylint: disable=protected-access
            self.device.sequencers[sequencer.identifier].sequence(qpysequence.todict())
            self.sequences[sequencer.identifier] = qpysequence

    def upload(self, port: str):  # TODO: check compatibility with QProgram
        """Upload all the previously compiled programs to its corresponding sequencers.

        This method must be called after the method ``compile`` in the compiler
        Args:
            port (str): The port of the sequencer to upload to.
        """
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=port)
        for sequencer in sequencers:
            seq_idx = sequencer.identifier
            # check is sequence has already been uploaded in a previous execution
            if seq_idx in self.sequences:
                # if the sequence was in the cache then it is to be run so we sync the sequencer to the others
                sequence = self.sequences[seq_idx]
                logger.info(
                    "Uploaded sequence program: \n %s", repr(sequence._program)  # pylint: disable=protected-access
                )
                self.device.sequencers[seq_idx].sequence(sequence.todict())
                self.device.sequencers[sequencer.identifier].sync_en(True)

    def _set_nco(self, sequencer_id: int):
        """Enable modulation of pulses and setup NCO frequency."""
        if self._get_sequencer_by_id(id=sequencer_id).hardware_modulation:
            self._set_hardware_modulation(
                value=self._get_sequencer_by_id(id=sequencer_id).hardware_modulation, sequencer_id=sequencer_id
            )
            self._set_frequency(
                value=self._get_sequencer_by_id(id=sequencer_id).intermediate_frequency, sequencer_id=sequencer_id
            )

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain_imbalance(self, value: float | str | bool, sequencer_id: int):
        """Set I and Q gain imbalance of sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """

        self._get_sequencer_by_id(id=sequencer_id).gain_imbalance = float(value)

        if self.is_device_active():
            self.device.sequencers[sequencer_id].mixer_corr_gain_ratio(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_phase_imbalance(self, value: float | str | bool, sequencer_id: int):
        """Set I and Q phase imbalance of sequencer.

         Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self._get_sequencer_by_id(id=sequencer_id).phase_imbalance = float(value)
        if self.is_device_active():
            self.device.sequencers[sequencer_id].mixer_corr_phase_offset_degree(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_markers(self, value: int, sequencer_id: int):
        """Set markers ON/OFF on qblox modules.

        For the RF modules, this command is also used to enable/disable:
            - The 2 outputs (for the QCM-RF).
            - The input and the output (for QRM-RF).

         Args:
            value (int): ON/OFF of the 4 markers in binary (range: 0-15 -> (0000)-(1111)). For the RF modules, the
                first 2 bits correspond to the ON/OFF value of the outputs/inputs and the last 2 bits correspond
                to the 2 markers.
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not int
        """
        self.device.sequencers[sequencer_id].marker_ovr_en(True)
        self.device.sequencers[sequencer_id].marker_ovr_value(value)

    def _map_connections(self):
        """Disable all connections and map sequencer paths with output channels."""
        # Disable all connections
        self.device.disconnect_outputs()

        for sequencer_dataclass in self.awg_sequencers:
            sequencer = self.device.sequencers[sequencer_dataclass.identifier]
            for path, output in zip(["I", "Q"], sequencer_dataclass.outputs):
                getattr(sequencer, f"connect_out{output}")(path)

    @property
    def out_offsets(self):
        """Returns the offsets of each output of the qblox module."""
        return self.settings.out_offsets

    def _get_sequencer_by_id(self, id: int):  # pylint: disable=redefined-builtin
        """Returns a sequencer with the given `id`."

        Args:
            id (int): Id of the sequencer.

        Raises:
            IndexError: There is no sequencer with the given `id`.

        Returns:
            AWGQbloxSequencer: Sequencer with the given `id`.
        """
        for sequencer in self.awg_sequencers:
            if sequencer.identifier == id:
                return sequencer
        raise IndexError(f"There is no sequencer with id={id}.")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PulseBusSchedule class."""
from __future__ import annotations

from bisect import insort
from dataclasses import dataclass, field

import numpy as np

from qililab.constants import PULSEBUSSCHEDULE
from qililab.pulse.pulse import Pulse
from qililab.pulse.pulse_event import PulseEvent
from qililab.utils import Waveforms


@dataclass
class PulseBusSchedule:
    """Container of Pulse objects addressed to the same bus.

    Args:
        port (str): Port (bus) alias.
        timeline (list[PulseEvent]): List of :class:`PulseEvent` objects the PulseBusSchedule is composed of.

    Examples:
        You can create a PulseBusSchedule targeting a bus or port in our chip by doing:

        .. code-block:: python3

            drag_pulse = Pulse(
                amplitude=1,
                phase=0.5,
                duration=200,
                frequency=1e9,
                pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            drag_pulse_event = PulseEvent(
                pulse=drag_pulse,
                start_time=0
            )

            drive_schedule = PulseBusSchedule(
                timeline=[drag_pulse_event],
                port="drive_q0"
            )

        You can add further pulse events to an already created PulseBusSchedule. To do so:

        .. code-block:: python3

            drag_pulse_0 = Pulse(
                amplitude=1,
                phase=0.5,
                duration=200,
                frequency=1e9,
                pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            drag_pulse_event = PulseEvent(
                pulse=drag_pulse_0,
                start_time=0
            )

            drive_schedule = PulseBusSchedule(
                timeline=[drag_pulse_event],
                port="drive_q0"
            )
            drag_pulse_1 = Pulse(
                amplitude=1,
                phase=0.5,
                duration=400,
                frequency=1e7,
                pulse_shape=Drag(num_sigmas=2, drag_coefficient=0.2)
            )
            drag_pulse_event_1 = PulseEvent(
                pulse=drag_pulse_1,
                start_time=204
            )
            drive_schedule.add_event(drag_pulse_event_1)
    """

    # FIXME: we may have one port being used by more than one bus. Use virtual ports instead.
    port: str  #: Port(bus) alias.
    timeline: list[PulseEvent] = field(
        default_factory=list
    )  #: List of PulseEvent objects the PulseBusSchedule is composed of.
    _pulses: set[Pulse] = field(init=False, default_factory=set)

    def __post_init__(self):
        """Sorts timeline and add used pulses to the pulses set if timeline is not empty."""
        if self.timeline:
            self.timeline.sort()
            for pulse_event in self.timeline:
                self._pulses.add(pulse_event.pulse)

    def add_event(self, pulse_event: PulseEvent):
        """Adds pulse event to sequence.

        Args:
            pulse_event (PulseEvent): :class:`PulseEvent` object.
        """
        self._pulses.add(pulse_event.pulse)
        insort(self.timeline, pulse_event)

    @property
    def end_time(self) -> int:
        """End of the PulseBusSchedule.

        Returns:
            int: End of the PulseBusSchedule."""
        end = 0
        for event in self.timeline:
            pulse_end = event.start_time + event.duration
            end = max(pulse_end, end)
        return end

    @property
    def start_time(self) -> int:
        """Start of the PulseBusSchedule.

        Returns:
            int: Start of the PulseBusSchedule."""
        return self.timeline[0].start_time

    @property
    def duration(self) -> int:
        """Duration of the PulseBusSchedule.

        Returns:
            int: Duration of the PulseBusSchedule."""
        return self.end_time - self.start_time

    @property
    def unique_pulses_duration(self) -> int:
        """Duration of the unique pulses, one immediately after the other.

        Returns:
            int: Duration of the unique pulses."""
        return sum(pulse.duration for pulse in self._pulses)

    @property
    def pulses(self):
        """Set of Pulse objects used in this PulseBusSchedule.

        Returns:
            str: The set of :class:`Pulse` objects."""
        return self._pulses

    @property
    def qubit(self):
        """Qubit index addressed by this PulseBusSchedule.

        Raises:
            ValueError: If more than one qubit is addressed by this PulseBusSchedule.

        Returns:
            int: The qubit index addressed by this PulseBusSchedule.
        """
        qubits = {pulse_event.qubit for pulse_event in self.timeline}
        if len(qubits) > 1:
            raise ValueError("More than one qubit is addressed by this PulseBusSchedule.")
        return qubits.pop()

    def __iter__(self):
        """Redirects __iter__ magic method."""
        return self.timeline.__iter__()

    def waveforms(self, resolution: float = 1.0, modulation: bool = True) -> Waveforms:
        """PulseBusSchedule 'waveforms' property.

        Args:
            resolution (float): The resolution of the pulses in ns.

        Returns:
            Waveforms: Class containing the I, Q waveforms for a specific qubit.
        """
        waveforms = Waveforms()
        time = 0
        for pulse_event in self.timeline:
            wait_time = round((pulse_event.start_time - time) / resolution)
            if wait_time > 0:
                waveforms.add(imod=np.zeros(shape=wait_time), qmod=np.zeros(shape=wait_time))
            time += wait_time
            if modulation:
                pulse_waveforms = pulse_event.modulated_waveforms(resolution=resolution)
            else:
                real_env = np.real(pulse_event.envelope(resolution=resolution))
                imag_env = np.imag(pulse_event.envelope(resolution=resolution))
                pulse_waveforms = Waveforms(i=real_env, q=imag_env)
            waveforms += pulse_waveforms
            time += pulse_event.duration

        return waveforms

    def qubit_schedules(self) -> list[PulseBusSchedule]:
        """Separates all the :class:`PulseEvent` objects that act on different qubits, and returns a list
        of PulseBusSchedule objects, each one acting on a single qubit.

        Returns:
            list[PulseBusSchedule]: List of PulseBusSchedule objects, each one acting on a single qubit.
        """
        schedules = []
        qubits = {pulse_event.qubit for pulse_event in self.timeline}
        for qubit in qubits:
            schedule = PulseBusSchedule(
                port=self.port, timeline=[pulse_event for pulse_event in self.timeline if pulse_event.qubit == qubit]
            )
            schedules.append(schedule)
        return schedules

    def to_dict(self):
        """Returns dictionary representation of the class.

        Returns:
            dict: Dictionary representation of the class.
        """
        return {
            PULSEBUSSCHEDULE.TIMELINE: [pulse_event.to_dict() for pulse_event in self.timeline],
            PULSEBUSSCHEDULE.PORT: self.port,
        }

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Loads PulseBusSchedule object from dictionary.

        Args:
            dictionary (dict): Dictionary representation of the PulseBusSchedule object.

        Returns:
            PulseBusSchedule: Loaded class.
        """
        timeline = [PulseEvent.from_dict(event) for event in dictionary[PULSEBUSSCHEDULE.TIMELINE]]
        port = dictionary[PULSEBUSSCHEDULE.PORT]
        return PulseBusSchedule(timeline=timeline, port=port)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the qcodes drivers supported by qililab, together with the interfaces for each
type of driver.

.. currentmodule:: qililab.drivers

Drivers
~~~~~~~

QBlox
======

.. autosummary::
    :toctree: api

    ~Cluster
    ~Pulsar
    ~SpiRack

Rohde & Schwarz
===============

.. autosummary::
    :toctree: api

    ~RhodeSchwarzSGS100A

Keithley
========

.. autosummary::
    :toctree: api

    ~Keithley2600

Yokogawa
========

.. autosummary::
    :toctree: api

    ~GS200

Interfaces
~~~~~~~~~~

.. currentmodule:: qililab.drivers.interfaces

.. autosummary::
    :toctree: api

    ~AWG
    ~Digitiser
    ~LocalOscillator
    ~CurrentSource
    ~VoltageSource
    ~Attenuator
"""
from .instruments import GS200, Cluster, ERASynthPlus, Keithley2600, Pulsar, RhodeSchwarzSGS100A, SpiRack



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keithley2600 & Keithley2600Channel drivers."""
from qcodes.instrument.channel import ChannelTuple, InstrumentModule
from qcodes.instrument_drivers.Keithley._Keithley_2600 import Keithley2600 as QCodesKeithley2600
from qcodes.instrument_drivers.Keithley._Keithley_2600 import Keithley2600Channel as QCodesKeithley2600Channel

from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import BaseInstrument, CurrentSource, VoltageSource


@InstrumentDriverFactory.register
class Keithley2600(QCodesKeithley2600, BaseInstrument):
    """
    This is the driver for the Keithley_2600 Source-Meter series,tested with Keithley_2614B

    Args:
        name (str): Name to use internally in QCoDeS
        address (str): VISA resource address
    """

    def __init__(self, name: str, address: str, **kwargs):
        super().__init__(name, address, **kwargs)
        self.submodules: dict[str, InstrumentModule | ChannelTuple] = {}  # resetting superclass submodules
        self.instrument_modules: dict[str, InstrumentModule] = {}  # resetting superclass instrument modules
        self._channel_lists: dict[str, ChannelTuple] = {}  # resetting superclass instrument channel lists
        self.channels: list[QCodesKeithley2600Channel] = []  # resetting superclass instrument channels
        # Add the channel to the instrument
        for ch in ["a", "b"]:
            ch_name = f"smu{ch}"
            channel = Keithley2600Channel(self, name=ch_name, channel=ch_name)
            self.add_submodule(ch_name, channel)
            self.channels.append(channel)

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name


class Keithley2600Channel(QCodesKeithley2600Channel, VoltageSource, CurrentSource):
    """
    Class to hold the two Keithley channels, i.e. SMUA and SMUB.

    It inherits from QCodes driver with extra on/off functionalities.

    Args:
        parent (QCodes.Instrument): The Instrument instance to which the channel is to be attached.
        name (str): The 'colloquial' name of the channel
        channel (str): The name used by the Keithley, i.e. either 'smua' or 'smub'
    """

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name

    def on(self) -> None:
        """Turn output on"""
        self.set(param_name="output", value="on")

    def off(self) -> None:
        """Turn output off"""
        self.set(param_name="output", value="off")



import functools

from .variable import Domain, Variable


def requires_domain(parameter: str, domain: Domain):
    """Decorator to denote that a parameter requires a variable of a specific domain.

    Args:
        parameter (str): The parameter name.
        domain (Domain): The variable's domain.
    """

    def decorator_function(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Get the argument by name
            param_value = (
                kwargs.get(parameter) if parameter in kwargs else args[1]
            )  # assuming 1 is the default position

            if isinstance(param_value, Variable) and param_value.domain != domain:
                raise ValueError(f"Expected domain {domain} for {parameter}, but got {param_value.domain}")
            return func(*args, **kwargs)

        return wrapper

    return decorator_function



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import numpy as np

from qililab.qprogram.blocks import Average, Block, ForLoop, InfiniteLoop, Loop, Parallel
from qililab.qprogram.decorators import requires_domain
from qililab.qprogram.operations import (
    Acquire,
    Measure,
    Play,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.variable import Domain, FloatVariable, IntVariable, ValueSource, Variable
from qililab.utils import DictSerializable
from qililab.waveforms import IQPair, Waveform


class QProgram(DictSerializable):
    """QProgram is a hardware-agnostic pulse-level programming interface for describing quantum programs.

    This class provides an interface for building quantum programs,
    including defining operations, managing variables, and handling blocks.
    It contains methods for creating, manipulating and controlling
    the execution flow of quantum operations within a program.

    Attributes:
        disable_autosync (bool): If set to True, then loop iterations are not automatically synced.

    Examples:

        The following example illustrates how to define a Rabi sequence using QProgram.

        .. code-block:: python3

            from qililab import QProgram, Domain, IQPair, Square

            qp = QProgram()

            # Pulse used for changing the state of qubit
            control_wf = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4.0, drag_correction=-2.5)

            # Pulse used for exciting the resonator for readout
            readout_wf = IQPair(I=Square(amplitude=1.0, duration=400), Q=Square(amplitude=0.0, duration=400))

            # Weights used during integration
            weights = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=1.0, duration=2000))

            # Declare a variable
            gain = qp.variable(Domain.Voltage)

            # Loop the variable's value over the range [0.0, 1.0]
            with qp.for_loop(variable=gain, start=0.0, stop=1.0, step=0.01):
                # Change the gain output of the drive_bus
                qp.set_gain(bus="drive_bus", gain=gain)

                # Play the control pulse
                qp.play(bus="drive_bus", waveform=control_wf)

                # Sync the buses
                qp.sync()

                # Play the readout pulse
                qp.play(bus="readout_bus", waveform=readout_wf, wait_time=120)

                # Acquire results
                qp.acquire(bus="readout_bus", weights=weights)

    """

    def __init__(self, disable_autosync: bool = False) -> None:
        self.disable_autosync: bool = disable_autosync

        self._body: Block = Block()
        self._buses: set[str] = set()
        self._variables: list[Variable] = []
        self._block_stack: deque[Block] = deque([self._body])

    def _append_to_block_stack(self, block: Block):
        self._block_stack.append(block)

    def _pop_from_block_stack(self):
        return self._block_stack.pop()

    @property
    def body(self) -> Block:
        """Get the body of the QProgram

        Returns:
            Block: The block of the body
        """
        return self._body

    @property
    def buses(self) -> set[str]:
        """Get the buses of the QProgram

        Returns:
            set[str]: A set of the names of the buses
        """
        return self._buses

    @property
    def variables(self) -> list[Variable]:
        """Get the variables

        Returns:
            list[Variable]: A list of variables
        """
        return self._variables

    @property
    def _active_block(self) -> Block:
        return self._block_stack[-1]

    def block(self):
        """Define a generic block for scoping operations.

        Blocks need to open a scope.

        Returns:
            Block: The block.

        Examples:

            >>> with qp.block() as block:
            >>>    # operations that shall be executed in the block
        """
        return QProgram._BlockContext(qprogram=self)

    def parallel(self, loops: list[Loop | ForLoop]):
        """Define a block for running multiple loops in parallel.

        Blocks need to open a scope.

        Examples:
            >>> gain = qp.variable(float)
            >>> frequency = qp.variable(float)
            >>> with qp.parallel(loops=[ForLoop(variable=frequency, start=0, stop=100, step=10),
                                        ForLoop(variable=gain, start=0.0, stop=1.0, step=0.1)]):
            >>>    # operations that shall be executed in the block

        Args:
            loops (list[Loop  |  ForLoop]): The loops to run in parallel

        Returns:
            Parallel: The parallel block.
        """
        return QProgram._ParallelContext(qprogram=self, loops=loops)

    def average(self, shots: int):
        """Define an acquire loop block with averaging in real time.

        Blocks need to open a scope.

        Args:
            iterations (int): The number of acquire iterations.

        Returns:
            AcquireLoop: The acquire_loop block.

        Examples:

            >>> with qp.acquire_loop(iterations=1000):
            >>>    # operations that shall be executed in the acquire_loop block
        """
        return QProgram._AverageContext(qprogram=self, shots=shots)

    def infinite_loop(self):
        """Define an infinite loop.

        Blocks need to open a scope.

        Examples:

            >>> with qp.infinite_loop():
            >>>    # operations that shall be executed in the infinite loop block

        Returns:
            InfiniteLoop: The infinite loop block.
        """
        return QProgram._InfiniteLoopContext(qprogram=self)

    def loop(self, variable: Variable, values: np.ndarray):
        """Define a loop block to iterate values over a variable.

        Blocks need to open a scope.

        Args:
            variable (Variable): The variable to be affected from the loop.
            values (np.ndarray): The values to iterate over.

        Returns:
            Loop: The loop block.

        Examples:

            >>> variable = qp.variable(int)
            >>> with qp.loop(variable=variable, values=np.array(range(100))):
            >>>    # operations that shall be executed in the loop block
        """

        return QProgram._LoopContext(qprogram=self, variable=variable, values=values)

    def for_loop(self, variable: Variable, start: int | float, stop: int | float, step: int | float = 1):
        """Define a for_loop block to iterate values over a variable.

        Blocks need to open a scope.

        Args:
            variable (Variable): The variable to be affected from the loop.
            start (int | float): The start value.
            stop (int | float): The stop value.
            step (int | float, optional): The step value. Defaults to 1.

        Returns:
            Loop: The loop block.

        Examples:

            >>> variable = qp.variable(int)
            >>> with qp.for_loop(variable=variable, start=0, stop=100, step=5)):
            >>>    # operations that shall be executed in the for_loop block
        """

        return QProgram._ForLoopContext(qprogram=self, variable=variable, start=start, stop=stop, step=step)

    def play(self, bus: str, waveform: Waveform | IQPair, wait_time: int | None = None):
        """Play a single waveform or an I/Q pair of waveforms on the bus.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (Waveform | IQPair): A single waveform or an I/Q pair of waveforms
        """
        operation = Play(bus=bus, waveform=waveform, wait_time=wait_time)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("duration", Domain.Time)
    def wait(self, bus: str, duration: int):
        """Adds a delay on the bus with a specified time.

        Args:
            bus (str): Unique identifier of the bus.
            time (int): Duration of the delay.
        """
        operation = Wait(bus=bus, duration=duration)
        self._active_block.append(operation)
        self._buses.add(bus)

    def acquire(self, bus: str, weights: IQPair):
        """Acquire results based on the given weights.

        Args:
            bus (str): Unique identifier of the bus.
            weights (IQPair): Weights used during acquisition.
        """
        operation = Acquire(bus=bus, weights=weights)
        self._active_block.append(operation)
        self._buses.add(bus)

    def measure(
        self,
        bus: str,
        waveform: IQPair,
        weights: IQPair | tuple[IQPair, IQPair] | tuple[IQPair, IQPair, IQPair, IQPair] | None = None,
        demodulation: bool = True,
        save_raw_adc: bool = False,
    ):
        """Play a pulse and acquire results.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (IQPair): Waveform played during measurement.
            weights (IQPair | tuple[IQPair, IQPair] | tuple[IQPair, IQPair, IQPair, IQPair] | None, optional): Weights used during acquisition. Defaults to None.
            demodulation (bool, optional): If demodulation is enabled. Defaults to True.
            save_raw_adc (bool, optional): If raw adc data should be saved. Defaults to True.
        """
        operation = Measure(
            bus=bus, waveform=waveform, weights=weights, demodulation=demodulation, save_raw_adc=save_raw_adc
        )
        self._active_block.append(operation)
        self._buses.add(bus)

    def sync(self, buses: list[str] | None = None):
        """Synchronize operations between buses, so the operations following will start at the same time.

        If no buses are given, then the synchronization will involve all buses present in the QProgram.

        Args:
            buses (list[str] | None, optional): List of unique identifiers of the buses. Defaults to None.
        """
        operation = Sync(buses=buses)
        self._active_block.append(operation)
        if buses:
            self._buses.update(buses)

    def reset_phase(self, bus: str):
        """Reset the absolute phase of the NCO associated with the bus.

        Args:
            bus (str): Unique identifier of the bus.
        """
        operation = ResetPhase(bus=bus)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("phase", Domain.Phase)
    def set_phase(self, bus: str, phase: float):
        """Set the absolute phase of the NCO associated with the bus.

        Args:
            bus (str): Unique identifier of the bus.
            phase (float): The new absolute phase of the NCO.
        """
        operation = SetPhase(bus=bus, phase=phase)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("frequency", Domain.Frequency)
    def set_frequency(self, bus: str, frequency: float):
        """Set the frequency of the NCO associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            frequency (float): The new frequency of the NCO.
        """
        operation = SetFrequency(bus=bus, frequency=frequency)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("gain", Domain.Voltage)
    def set_gain(self, bus: str, gain: float):
        """Set the gain of the AWG associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            gain (float): The new gain of the AWG.
        """
        operation = SetGain(bus=bus, gain=gain)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("offset_path0", Domain.Voltage)
    @requires_domain("offset_path1", Domain.Voltage)
    def set_offset(self, bus: str, offset_path0: float, offset_path1: float):
        """Set the gain of the AWG associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            offset_path0 (float): The new offset of the AWG for path0.
            offset_path1 (float): The new offset of the AWG for path1.
        """
        operation = SetOffset(bus=bus, offset_path0=offset_path0, offset_path1=offset_path1)
        self._active_block.append(operation)
        self._buses.add(bus)

    def variable(self, domain: Domain, type: type[int | float] | None = None):  # pylint: disable=redefined-builtin
        """Declare a variable.

        Args:
            type (int | float): The type of the variable.

        Raises:
            NotImplementedError: If an unsupported type is provided.

        Returns:
            IntVariable | FloatVariable: The variable.
        """

        def _int_variable(domain: Domain = Domain.Scalar) -> IntVariable:
            variable = IntVariable(domain)
            self._variables.append(variable)
            return variable

        def _float_variable(domain: Domain = Domain.Scalar) -> FloatVariable:
            variable = FloatVariable(domain)
            self._variables.append(variable)
            return variable

        if domain is Domain.Scalar and type is None:
            raise ValueError("You must specify a type in a scalar variable.")
        if domain is not Domain.Scalar and type is not None:
            raise ValueError("When declaring a variable of a specific domain, its type is inferred by its domain.")

        if domain is Domain.Scalar:
            if type == int:
                return _int_variable(domain)
            if type == float:
                return _float_variable(domain)

        if domain == Domain.Time:
            return _int_variable(domain)
        if domain in [Domain.Frequency, Domain.Phase, Domain.Voltage]:
            return _float_variable(domain)
        raise NotImplementedError

    class _BlockContext:
        def __init__(self, qprogram: "QProgram"):
            self.qprogram = qprogram
            self.block: Block = Block()

        def __enter__(self):
            self.qprogram._append_to_block_stack(block=self.block)
            return self.block

        def __exit__(self, exc_type, exc_value, exc_tb):
            block = self.qprogram._pop_from_block_stack()
            self.qprogram._active_block.append(block)

    class _InfiniteLoopContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(self, qprogram: "QProgram"):  # pylint: disable=super-init-not-called
            self.qprogram = qprogram
            self.block: InfiniteLoop = InfiniteLoop()

    class _ParallelContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(self, qprogram: "QProgram", loops: list[Loop | ForLoop]):  # pylint: disable=super-init-not-called
            self.qprogram = qprogram
            self.block: Parallel = Parallel(loops=loops)

        def __enter__(self) -> Parallel:
            for loop in self.block.loops:
                loop.variable._source = ValueSource.Dependent
                loop.variable._value = None
            return super().__enter__()

        def __exit__(self, exc_type, exc_value, exc_tb):
            for loop in self.block.loops:
                loop.variable._source = ValueSource.Free
            super().__exit__(exc_type, exc_value, exc_tb)

    class _LoopContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(  # pylint: disable=super-init-not-called
            self, qprogram: "QProgram", variable: Variable, values: np.ndarray
        ):
            self.qprogram = qprogram
            self.block: Loop = Loop(variable=variable, values=values)

        def __enter__(self) -> Loop:
            self.block.variable._source = ValueSource.Dependent
            self.block.variable._value = None
            return super().__enter__()

        def __exit__(self, exc_type, exc_value, exc_tb):
            self.block.variable._source = ValueSource.Free
            super().__exit__(exc_type, exc_value, exc_tb)

    class _ForLoopContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(  # pylint: disable=super-init-not-called
            self, qprogram: "QProgram", variable: Variable, start: int | float, stop: int | float, step: int | float
        ):
            self.qprogram = qprogram
            self.block: ForLoop = ForLoop(variable=variable, start=start, stop=stop, step=step)

        def __enter__(self) -> ForLoop:
            self.block.variable._source = ValueSource.Dependent
            self.block.variable._value = None
            return super().__enter__()

        def __exit__(self, exc_type, exc_value, exc_tb):
            self.block.variable._source = ValueSource.Free
            super().__exit__(exc_type, exc_value, exc_tb)

    class _AverageContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(self, qprogram: "QProgram", shots: int):  # pylint: disable=super-init-not-called
            self.qprogram = qprogram
            self.block: Average = Average(shots=shots)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the code of the sequencer of the Qblox QRM."""
from qcodes import Instrument
from qcodes import validators as vals
from qpysequence.acquisitions import Acquisitions
from qpysequence.program import Loop, Program, Register
from qpysequence.program.instructions import Acquire, AcquireWeighed, Move
from qpysequence.weights import Weights

from qililab.config import logger
from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import Digitiser
from qililab.result.qblox_results.qblox_result import QbloxResult

from .sequencer_qcm import SequencerQCM


@InstrumentDriverFactory.register
class SequencerQRM(SequencerQCM, Digitiser):
    """Qililab's driver for QBlox-instruments digitiser Sequencer

    Args:
        parent (Instrument): Parent for the sequencer instance.
        name (str): Sequencer name
        seq_idx (int): sequencer identifier index
        sequence_timeout (int): timeout to retrieve sequencer state in minutes
        acquisition_timeout (int): timeout to retrieve acquisition state in minutes
    """

    def __init__(self, parent: Instrument, name: str, seq_idx: int):
        super().__init__(parent=parent, name=name, seq_idx=seq_idx)
        self.add_parameter(name="sequence_timeout", vals=vals.Ints(), set_cmd=None, initial_value=1)
        self.add_parameter(name="acquisition_timeout", vals=vals.Ints(), set_cmd=None, initial_value=1)
        self.add_parameter(name="weights_i", vals=vals.Lists(), set_cmd=None, initial_value=[])
        self.add_parameter(name="weights_q", vals=vals.Lists(), set_cmd=None, initial_value=[])
        self.add_parameter(name="weighed_acq_enabled", vals=vals.Bool(), set_cmd=None, initial_value=False)

    def get_results(self) -> QbloxResult:
        """Wait for sequencer to finish sequence, wait for acquisition to finish and get the acquisition results.
        If any of the timeouts is reached, a TimeoutError is raised.

        Returns:
            QbloxResult: Class containing the acquisition results.

        """
        flags = self.parent.get_sequencer_state(sequencer=self.seq_idx, timeout=self.get("sequence_timeout"))
        logger.info("Sequencer[%d] flags: \n%s", self.seq_idx, flags)
        self.parent.get_acquisition_state(sequencer=self.seq_idx, timeout=self.get("acquisition_timeout"))
        if self.parent.get("scope_acq_sequencer_select") == self.seq_idx:
            self.parent.store_scope_acquisition(sequencer=self.seq_idx, name="default")
        result = self.parent.get_acquisitions(sequencer=self.seq_idx)["default"]["acquisition"]

        integration_length = (
            len(self.get("weights_i")) if self.get("weighed_acq_enabled") else self.get("integration_length_acq")
        )

        self.set("sync_en", False)
        return QbloxResult(integration_lengths=[integration_length], qblox_raw_results=[result])

    def _init_weights_registers(self, registers: tuple[Register, Register], values: tuple[int, int], program: Program):
        """Initialize the weights `registers` to the `values` specified and place the required instructions in the
        setup block of the `program`."""
        move_0 = Move(values[0], registers[0])
        move_1 = Move(values[1], registers[1])
        setup_block = program.get_block(name="setup")
        setup_block.append_components([move_0, move_1], bot_position=1)

    def _generate_acquisitions(self, num_bins: int) -> Acquisitions:
        """Generate Acquisitions object, currently containing a single acquisition named "default" with
        index = 0.

        Args:
            num_bins (int): number of bins

        Returns:
            Acquisitions: Acquisitions object.
        """
        acquisitions = Acquisitions()
        acquisitions.add(name="default", num_bins=num_bins, index=0)

        return acquisitions

    def _generate_weights(self) -> Weights:
        """Generate acquisition weights.

        Returns:
            Weights: Acquisition weights.
        """
        weights = Weights()
        if len(self.get("weights_i")) > 0 and len(self.get("weights_q")) > 0:
            pair = (self.get("weights_i"), self.get("weights_q"))
            weights.add_pair(pair=pair, indices=(0, 1))
        return weights

    def _append_acquire_instruction(
        self, loop: Loop, bin_index: Register | int, weight_regs: tuple[Register, Register]
    ):
        """Append an acquire instruction to the loop."""
        weighed_acq = self.get("weighed_acq_enabled")

        acq_instruction = (
            AcquireWeighed(
                acq_index=0,
                bin_index=bin_index,
                weight_index_0=weight_regs[0],
                weight_index_1=weight_regs[1],
                wait_time=self._MIN_WAIT_TIME,
            )
            if weighed_acq
            else Acquire(acq_index=0, bin_index=bin_index, wait_time=self._MIN_WAIT_TIME)
        )
        loop.append_component(acq_instruction)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qblox_instruments.qcodes_drivers import Cluster as QcodesCluster
from qblox_instruments.qcodes_drivers.qcm_qrm import QcmQrm as QcodesQcmQrm
from qcodes import Instrument
from qcodes.instrument import DelegateParameter
from qcodes.instrument.channel import ChannelTuple, InstrumentModule

from qililab.drivers import parameters
from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import Attenuator, BaseInstrument, LocalOscillator

from .sequencer_qcm import SequencerQCM
from .sequencer_qrm import SequencerQRM


@InstrumentDriverFactory.register
class Cluster(QcodesCluster, BaseInstrument):  # pylint: disable=abstract-method
    """Qililab's driver for QBlox-instruments Cluster.

    Args:
        name (str): The name/alias of the cluster.
        address (str): The IP address of the cluster.

    Keyword Args:
        **kwargs: Any additional keyword arguments provided are passed to its `parent class
            <https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/api_reference/cluster.html>`_.
    """

    def __init__(self, name: str, address: str | None = None, **kwargs):
        super().__init__(name, identifier=address, **kwargs)

        # registering only the slots specified in the dummy config if that is the case
        if "dummy_cfg" in kwargs:
            slot_ids = list(kwargs["dummy_cfg"].keys())
        else:
            slot_ids = list(range(1, self._num_slots + 1))

        # Save information about modules actually being present in the cluster
        old_submodules = self.submodules
        submodules_present = [submodule.get("present") for submodule in old_submodules.values()]

        # Add qcm-qrm's to the cluster
        self.submodules: dict[str, InstrumentModule | ChannelTuple] = {}  # resetting superclass submodules
        self.instrument_modules: dict[str, InstrumentModule] = {}  # resetting superclass instrument modules
        self._channel_lists: dict[str, ChannelTuple] = {}  # resetting superclass channel lists

        for slot_idx in slot_ids:
            if submodules_present[slot_idx - 1]:
                module = QcmQrm(self, f"module{slot_idx}", slot_idx)
                self.add_submodule(f"module{slot_idx}", module)
            else:
                old_module = old_submodules[f"module{slot_idx}"]
                self.add_submodule(f"module{slot_idx}", old_module)

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name


class QcmQrm(QcodesQcmQrm, BaseInstrument):
    """Qililab's driver for QBlox-instruments QcmQrm

    Args:
        parent (Instrument): Instrument´s parent
        name (str): Name of the instrument
        slot_idx (int): Index of the slot
    """

    def __init__(self, parent: Instrument, name: str, slot_idx: int):
        super().__init__(parent, name, slot_idx)

        # Add sequencers
        self.submodules: dict[str, InstrumentModule | ChannelTuple] = {}  # resetting superclass submodules
        self.instrument_modules: dict[str, InstrumentModule] = {}  # resetting superclass instrument modules
        self._channel_lists: dict[str, ChannelTuple] = {}  # resetting superclass channel lists
        sequencer_class = SequencerQCM if self.is_qcm_type else SequencerQRM
        for seq_idx in range(6):
            seq = sequencer_class(parent=self, name=f"sequencer{seq_idx}", seq_idx=seq_idx)  # type: ignore
            self.add_submodule(f"sequencer{seq_idx}", seq)

        # Add RF submodules
        if super().is_rf_type:
            # Add local oscillators
            # We use strings as channels to keep the in/out name and conserve the same
            # logic as for the attenuator or other instruments
            lo_channels = ["out0_in0"] if super().is_qrm_type else ["out0", "out1"]
            for channel in lo_channels:
                lo = QcmQrmRfLo(name=f"{name}_lo_{channel}", parent=self, channel=channel)
                self.add_submodule(f"{name}_lo_{channel}", lo)

            # Add attenuators
            att_channels = ["out0", "in0"] if super().is_qrm_type else ["out0", "out1"]
            for channel in att_channels:
                att = QcmQrmRfAtt(name=f"{name}_attenuator_{channel}", parent=self, channel=channel)
                self.add_submodule(f"{name}_attenuator_{channel}", att)

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name


class QcmQrmRfLo(InstrumentModule, LocalOscillator):
    """LO driver for the QCM / QRM - RF instrument
    Set and get methods from InstrumentModule override LocalOscillator's
    """

    def __init__(self, name: str, parent: QcmQrm, channel: str, **kwargs):
        super().__init__(parent, name, **kwargs)
        self.device = parent

        # get the name of the parameter for the lo frequency
        lo_frequency = parameters.lo.frequency
        self.add_parameter(
            lo_frequency,
            label="Delegated parameter for local oscillator frequency",
            source=parent.parameters[f"{channel}_lo_freq"],
            parameter_class=DelegateParameter,
        )
        self.add_parameter(
            "status",
            label="Delegated parameter device status",
            source=parent.parameters[f"{channel}_lo_en"],
            parameter_class=DelegateParameter,
        )

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name

    def on(self):
        self.set("status", True)

    def off(self):
        self.set("status", False)


class QcmQrmRfAtt(InstrumentModule, Attenuator):
    """Attenuator driver for the QCM / QRM - RF instrument
    Set and get methods from InstrumentModule override Attenuator's
    """

    def __init__(self, name: str, parent: QcmQrm, channel: str, **kwargs):
        super().__init__(parent, name, **kwargs)
        self.device = parent

        # get the name of the parameter for the attenuation
        attenuation = parameters.attenuator.attenuation
        self.add_parameter(
            attenuation,
            label="Delegated parameter for attenuation",
            source=parent.parameters[f"{channel}_att"],
            parameter_class=DelegateParameter,
        )

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module containing the interface for the voltage_sources """
from abc import abstractmethod

from .base_instrument import BaseInstrument
from .instrument_interface_factory import InstrumentInterfaceFactory


@InstrumentInterfaceFactory.register
class VoltageSource(BaseInstrument):
    """Voltage source interface with set, get, on & off abstract methods"""

    @abstractmethod
    def on(self) -> None:
        """Start VoltageSource output"""

    @abstractmethod
    def off(self) -> None:
        """Stop VoltageSource outout"""



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" AWG Sequencer """
from dataclasses import asdict, dataclass

from qililab.utils.asdict_factory import dict_factory


@dataclass
class AWGSequencer:  # pylint: disable=too-many-instance-attributes
    """AWG Sequencer

    Args:
        identifier (int): The identifier of the sequencer
        chip_port_id (str | None): Port identifier of the chip where a specific sequencer is connected to.
                                    By default, using the first sequencer
        outputs (list(int)): List of integers containing the outputs of the I and Q paths respectively. If the list only
            contains one item, then only one output will be connected.
        output_q (int): AWG output associated with the Q channel of the sequencer
        intermediate_frequency (float): Frequency for each sequencer
        gain_imbalance (float): Amplitude added to the Q channel.
        phase_imbalance (float): Dephasing.
        hardware_modulation  (bool): Flag to determine if the modulation is performed by the device
        gain_i (float): Gain step used by the I channel of the sequencer.
        gain_q (float): Gain step used by the Q channel of the sequencer.
        offset_i (float): I offset (unitless). amplitude + offset should be in range [0 to 1].
        offset_q (float): Q offset (unitless). amplitude + offset should be in range [0 to 1].
    """

    identifier: int
    chip_port_id: str | None
    # list containing the outputs for the I and Q paths e.g. [3, 2] means I path is connected to output 3 and Q path is connected to output 2
    outputs: list[int]
    intermediate_frequency: float
    gain_imbalance: float | None
    phase_imbalance: float | None
    hardware_modulation: bool
    gain_i: float
    gain_q: float
    offset_i: float
    offset_q: float

    def to_dict(self):
        """Return a dict representation of an AWG Sequencer."""
        return asdict(self, dict_factory=dict_factory) | {"outputs": self.outputs}



from .measurement_result import MeasurementResult
from .qblox_measurement_result import QbloxMeasurementResult
from .qprogram_results import QProgramResults
from .quantum_machines_measurement_result import QuantumMachinesMeasurementResult
